package cs;

import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;

public class BSCCreditExtracter extends DefaultCreditExtracter {
	private static final String TEAM_NAME = "Team China";
	private static final String URL_TEMPLATE =
			"https://boincstats.com/en/stats/challenge/user/list/{0}/credits/0/";

	private Map<String, User> userMap;

	@Override
	public void doExtractCredit(final Project project) throws Exception {
		System.out.printf("提取%s项目的用户积分...\n", project.name);
		this.initUserMap(project);
		// extract credit
		this.extract(project);
		// new line
		System.out.println();
	}

	private void initUserMap(final Project project) {
		this.userMap = new HashMap<String, User>(project.userList.size() * 100 / 75);
		for (User user : project.userList) {
			this.userMap.put(user.boincName.toLowerCase(), user);
		}
	}

	private void extract(final Project project) throws Exception {
		final String url = MessageFormat.format(URL_TEMPLATE, project.challenge.bscid);
		for (int i = 0;; i++) {
			final int index = 100 * i;
			boolean end = ExceptionHandler.retry(new ExceptionHandler.Handler<Boolean>() {
				@Override
				public Boolean handle() throws Exception {
					String html = UrlFetchUtil.get(url + index);
					if (html == null) {
						throw new RuntimeException(
								"Network error! 'User stats on credits' html is null("
										+ url
										+ ").");
					}
					Document doc = Jsoup.parse(html);
					for (int end = index + 100, i = index; i++ < end;) {
						Element tr = doc.getElementById("rank_" + i);
						if (tr == null) {
							// System.out.println("find finish! index:" + index);
							return true;
						}
						String teamName = tr.child(1).html()
								.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
								.replaceFirst("(?:\\&nbsp;|\\s*)$", "");
						if (!TEAM_NAME.equalsIgnoreCase(teamName)) {
							continue;
						}
						String userName = tr.child(0).html()
								.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
								.replaceFirst("(?:\\&nbsp;|\\s*)$", "");
						String credit = tr.child(3).ownText()
								.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
								.replaceFirst("(?:\\&nbsp;|\\s*)$", "")
								.replaceAll(",", "");
						if (userName == null || credit == null) {
							continue;
						}
						if (!BSCCreditExtracter.this.userMap.containsKey(userName.toLowerCase())) {
							continue;
						}
						System.out.printf("user: %s, credit: %s\n", userName, credit);
						User user = BSCCreditExtracter.this.userMap.remove(userName.toLowerCase());
						user.totalCredit = 0;
						user.credit = Double.parseDouble(credit);
						if (user.credit > 0) {
							project.memberNumber++;
							if (!project.challenge.isTeam) {
								project.totalCredit += user.credit;
							}
						}
						if (BSCCreditExtracter.this.userMap.isEmpty()) {
							return true;
						}
					}
					return false;
				}
			});
			if (end) {
				break;
			}
		}
	}

}
