package cs;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.ListIterator;
import java.util.Map;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

public class BSCreditExtracter extends DefaultCreditExtracter {

	private DateFormat dateFormat = new SimpleDateFormat("yy-MM-dd");

	private static final String URL_TEMPLATE_USER =
			"https://boincstats.com/en/stats/{0}/user/detail/{1}/lastDays";

	@Override
	public void doExtractCredit(final Project project) throws Exception {
		System.out.printf("提取%s项目的用户积分...\n", project.name);
		this.getUserId(project);
		// extract credit
		this.extract(project);
		// new line
		System.out.println();
	}

	private void loadUserId(final Project project, final Team team) throws Exception {
		if (this.fuckWCG(project, team)) {
			return;
		}
		ExceptionHandler.retry(new ExceptionHandler.Handler<Object>() {
			@Override
			public Object handle() throws Exception {
				String url = StringUtil.resolveURL(project.url,
						"./team_email_list.php?xml=1&teamid="
								+ team.id);
				String xml = UrlFetchUtil.get(url);
				if (xml == null) {
					throw new RuntimeException(
							"Network error! team_email_list xml is null. team id:"
									+ team.id);
				}
				InputStream is = new ByteArrayInputStream(xml.getBytes("UTF-8"));
				XMLEventReader reader = XMLInputFactory.newInstance()
						.createXMLEventReader(is, "UTF-8");

				StartElement root = XMLUtil.getRootElement(reader);
				if (root == null) {
					throw new RuntimeException("Invalid xml format, root element not found.");
				}
				if (root.getName().getLocalPart().equals("error")) {
					String msg = "team_email_list xml has error. team id:" + team.id;
					StartElement se = null;
					while ((se = XMLUtil.nextTag(reader)) != null) {
						String n = se.getName().getLocalPart();
						if (n.equals("error_string") || n.equals("error_msg")) {
							msg = reader.getElementText();
							break;
						}
					}
					throw new RuntimeException(msg);
				}

				Map<String, User> userNameMap = new HashMap<String, User>(
						project.userList.size() / 75 * 100);
				for (User user : project.userList) {
					if (user.team != team) {
						continue;
					}
					for (String name : user.boincName.split("/")) {
						userNameMap.put(name.toLowerCase(), user);
					}
				}
				while (XMLUtil.nextTag(reader, "user") != null) {
					String name = null, id = null, totalCredit = null, expavg_credit = null;
					while (reader.hasNext()) {
						XMLEvent e = reader.nextEvent();
						if (e.isStartElement()) {
							StartElement se = e.asStartElement();
							String qname = se.getName().getLocalPart();
							if (id == null && qname.equals("id")) {
								id = reader.getElementText();
							} else if (name == null && qname.equals("name")) {
								name = reader.getElementText();
							} else if (totalCredit == null && qname.equals("total_credit")) {
								totalCredit = reader.getElementText();
							} else if (expavg_credit == null && qname.equals("expavg_credit")) {
								expavg_credit = reader.getElementText();
							}
						} else if (e.isEndElement()) {
							EndElement ee = e.asEndElement();
							if (ee.getName().getLocalPart().equals("user")) {
								break;
							}
						}
						if (name != null && id != null && totalCredit != null
								&& expavg_credit != null) {
							break;
						}
					}
					if (id == null) {
						throw new RuntimeException("Invalid xml format, id not found");
					} else if (name == null) {
						throw new RuntimeException("Invalid xml format, name not found");
					} else if (totalCredit == null) {
						throw new RuntimeException("Invalid xml format, total_credit not found");
					}
					// 检查total_credit，以排除同名称的多个账户
					if (totalCredit.equals("0")) {
						continue;
					}
					double ec = Double.parseDouble(expavg_credit);
					User user = userNameMap.get(name.toLowerCase());
					if (user == null
							|| (user.id != null && user.expavg_credit >= ec)) {
						continue;
					}
					user.expavg_credit = ec;
					user.id = id;
				}
				reader.close();
				is.close();
				return null;
			}
		});

	}

	private void getUserId(final Project project) throws Exception {
		for (Team team : project.teamList) {
			if (Team.NOT_TEAM == team) {
				continue;
			}
			if (team.id == null) {
				team.id = this.getTeamId(team.name, project.url);
			}
			this.loadUserId(project, team);
		}
	}

	private void extract(final Project project) throws Exception {
		double totalCredit = 0;
		int memberNumber = 0;
		for (final User user : project.userList) {
			if (user.id == null) {
				continue;
			}
			ExceptionHandler.retry(new ExceptionHandler.Handler<Object>() {
				@Override
				public Object handle() throws Exception {
					String url = MessageFormat.format(URL_TEMPLATE_USER, project.id, user.id);
					String html = UrlFetchUtil.get(url);
					if (html == null) {
						throw new RuntimeException("Network error! 'user lastDays' html is null("
								+ url
								+ ").");
					}
					Document doc = Jsoup.parse(html);
					Elements trs = doc.select("#tblStats tr");

					boolean foundEndDay = false;
					int days = project.days;
					try {
						ListIterator<Element> itrs = trs.listIterator();
						Element tr = null;
						while (itrs.hasNext() && days > 0) {
							tr = itrs.next();
							if (!tr.hasClass("header")) {
								continue;
							}
							int colIndex = 1;
							if (!foundEndDay) {
								// find "end day" column index
								for (; colIndex <= 10; colIndex++) {
									String dateString = tr.child(colIndex).html()
											.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
											.replaceFirst("(?:\\&nbsp;|\\s*)$", "");
									Date date = this.parseDate(dateString);
									if (date.getTime() == project.endDate.getTime()) {
										// "Total credit" row
										Element totalCreditRow = trs.get(itrs.nextIndex());
										String totalCreditString = totalCreditRow.child(colIndex)
												.html()
												.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
												.replaceFirst("(?:\\&nbsp;|\\s*)$", "")
												.replaceAll(",", "");
										if (totalCreditString.equals("")) {
											// 新用户没积分
											days = 0;
										} else {
											user.totalCredit = Double
													.parseDouble(totalCreditString);
											foundEndDay = true;
										}
										break;
									}
								}
							}
							if (foundEndDay) {
								// "Total credit" row
								itrs.next();
								// "Credit per day" row
								tr = itrs.next();
								for (; colIndex <= 10 && days > 0; colIndex++) {
									String credit = tr.child(colIndex).html()
											.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
											.replaceFirst("(?:\\&nbsp;|\\s*)$", "")
											.replaceAll(",", "");
									user.credit += Double.parseDouble(credit);
									days--;
								}
							}
						}
					} catch (RuntimeException e) {
						// 用户ID不存在，通常为新注册用户，而boincstats还没更新
						if (html.contains("User ID doesn't exist")) {
							// next user
							return null;
						}
						throw e;
					}
					if (days > 0) {
						throw new RuntimeException("已超出提取时间范围");
					}
					System.out.printf("user: %s, credit: %s\n", user.boincName, user.credit);
					return null;
				}

				private Date parseDate(final String str) {
					Date d = null;
					try {
						d = BSCreditExtracter.this.dateFormat.parse(str);
					} catch (ParseException e) {
						throw new RuntimeException(e);
					}
					return d;
				}
			});
			if (user.credit > 0) {
				totalCredit += user.credit;
				memberNumber++;
			}
		}
		project.memberNumber = memberNumber;
		if (!project.challenge.isTeam) {
			project.totalCredit = totalCredit;
		}
	}

	private boolean fuckWCG(final Project project, final Team team) throws Exception {
		if (!project.url.contains("worldcommunitygrid")) {
			return false;
		}
		System.out.println("搜索各用户的ID(较耗时)...");
		int userNum = 0, userTotal = 0;
		Map<String, User> userNameMap = new HashMap<String, User>(project.userList.size() / 75 * 100);
		for (User user : project.userList) {
			if (user.team != team) {
				continue;
			}
			for (String name : user.boincName.split("/")) {
				userNameMap.put(name.toLowerCase(), user);
			}
			userTotal++;
		}
		String URL_TEMPLATE_USERLIST = "https://boincstats.com/en/stats/{0}/user/list/8/{1,number,#}/{2}";
		boolean isEnd = false;
		for (int i = 0; !isEnd; i += 100) {
			String url = MessageFormat.format(URL_TEMPLATE_USERLIST, project.id, i, team.id);
			String html = UrlFetchUtil.get(url);
			if (html == null) {
				throw new RuntimeException("Network error! 'WCG user list' html is null(" + url + ").");
			}
			Document doc = Jsoup.parse(html);
			for (int j = i + 1; j <= i + 100; j++) {
				Element tr = doc.getElementById("rank_" + j);
				if (null == tr) {
					isEnd = true;
					break;
				}
				String name = tr.child(4).text().trim().toLowerCase();
				if (userNameMap.containsKey(name)) {
					String id = tr.child(11).child(0).val();
					userNameMap.get(name).id = id;
					if (++userNum >= userTotal) {
						isEnd = true;
						break;
					}
				}
			}
		}
		// 处理没找到ID的情况
		if (userNum < userTotal) {
			StringBuilder sb = new StringBuilder();
			for (User user : userNameMap.values()) {
				if (null == user.id) {
					if (sb.length() > 0) {
						sb.append(", ");
					}
					sb.append(user.forumName).append("/").append(user.boincName);
				}
			}
			throw new RuntimeException("部分user(" + sb.toString() + ")没找到ID，请检查用户名正确与否。");
		}
		return true;
	}
}
