package cs;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.URLEncoder;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.events.StartElement;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.select.Elements;

public class DefaultCreditExtracter implements CreditExtracter {

	private static final String URL_TEMPLATE_TEAM =
			"https://boincstats.com/en/stats/{0}/team/detail/{1}";

	@Override
	public void extractCredit(final Project project) throws Exception {
		this.doExtractCredit(project);
		System.out.printf("提取%s项目的团队信息...\n", project.name);
		this.setTeamInfo(project);
		// new line
		System.out.println();
	}

	protected void doExtractCredit(final Project project) throws Exception {
		// NOOP, override only
	}

	private void setTeamInfo(final Project project) throws Exception {
		Set<Team> teamSet = new HashSet<Team>();
		int count = 0;
		for (User user : project.userList) {
			if (user.credit > 0) {
				count++;
				if (Team.NOT_TEAM == user.team) {
					continue;
				}
				user.team.memberNumber++;
				teamSet.add(user.team);
			}
		}
		project.memberNumber = count;
		project.teamList = new ArrayList<Team>(teamSet);

		if (!project.challenge.disableProjectBalanceFactor) {
			for (Team team : teamSet) {
				if (project.id == null) {// 非boinc项目
					/*
					 * 手动输入团队信息
					 */
					team.ranking = Integer.parseInt(Console.answerPattern(team.name + "团队在"
							+ project.name + "项目上的国际排名",
							"[123456789]\\d*"));
					team.memberTotalNumber = Integer.parseInt(Console.answerPattern(team.name
							+ "团队在" + project.name + "项目上的总成员数",
							"[123456789]\\d*"));
				} else {
					if (team.id == null) {
						team.id = this.getTeamId(team.name, project.url);
					}
					this.fetchTeamInfo(project, team);
				}
			}
		}
	}

	private void fetchTeamInfo(final Project project, final Team team) throws Exception {
		ExceptionHandler.retry(new ExceptionHandler.Handler<Object>() {
			@Override
			public Object handle() throws Exception {
				String url = MessageFormat.format(URL_TEMPLATE_TEAM, project.id, team.id);
				String html = UrlFetchUtil.get(url);
				if (html == null) {
					throw new RuntimeException("Network error! 'team detailed stats' html is null("
							+ url
							+ ").");
				}
				Document doc = Jsoup.parse(html);
				Elements trs = doc.select("#tblStats tr");
				String ranking = trs.get(6).child(1).child(0).html()
						.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
						.replaceFirst("(?:\\&nbsp;|\\s*)$", "").replaceAll(",", "");
				String memberNumber = trs.get(32).child(1).child(0).html()
						.replaceFirst("^(?:\\&nbsp;|\\s*)", "")
						.replaceFirst("(?:\\&nbsp;|\\s*)$", "").replaceAll(",", "");
				team.ranking = Integer.parseInt(ranking);
				team.memberTotalNumber = Integer.parseInt(memberNumber);
				return null;
			}
		});
	}

	protected String getTeamId(final String name, final String rpcUrl) throws Exception {
		String teamId = ExceptionHandler.retry(new ExceptionHandler.Handler<String>() {
			@Override
			public String handle() throws Exception {
				String team_name = URLEncoder.encode(name, "UTF-8");
				String url = "./team_lookup.php?format=xml&team_name=" + team_name;
				url = StringUtil.resolveURL(rpcUrl, url);
				String xml = UrlFetchUtil.get(url);
				if (xml == null) {
					throw new RuntimeException("Network error! team_lookup xml is null.");
				}
				InputStream is = new ByteArrayInputStream(xml.getBytes("UTF-8"));
				XMLEventReader reader = XMLInputFactory.newInstance()
						.createXMLEventReader(is, "UTF-8");

				StartElement root = XMLUtil.getRootElement(reader);
				if (root == null) {
					throw new RuntimeException("Invalid xml format, root element not found.");
				}
				if (root.getName().getLocalPart().equals("error")) {
					String msg = "team_lookup xml has error.";
					StartElement se = null;
					while ((se = XMLUtil.nextTag(reader)) != null) {
						String n = se.getName().getLocalPart();
						if (n.equals("error_string") || n.equals("error_msg")) {
							msg = reader.getElementText();
							break;
						}
					}
					throw new RuntimeException(msg);
				}
				String teamId = null;
				while (teamId == null && XMLUtil.nextTag(reader, "team") != null) {
					String id = null;
					if (XMLUtil.nextTag(reader, "id") != null) {
						id = reader.getElementText();
					}
					if (XMLUtil.nextTag(reader, "name") != null) {
						if (reader.getElementText().equalsIgnoreCase(name)) {
							teamId = id;
						}
					}
				}
				if (teamId == null) {
					throw new RuntimeException("无法找到团队" + name + "的id，可能是该团队在该项目下还没建立或团队名称有误。");
				}

				reader.close();
				is.close();
				return teamId;
			}
		});
		return teamId;
	}
}
