package cs;

public class ExceptionHandler {

	private static String DEFAULT_RETRY_COUNT = "6";

	public interface Handler<T> {
		public T handle() throws Exception;
	}

	private static int getRetryCount() {
		int retryCount = Integer.parseInt(System.getProperty("retry", DEFAULT_RETRY_COUNT));
		return retryCount < 1 ? 1 : retryCount;
	}

	public static <T extends Object> T retry(final Handler<T> handler) throws Exception {
		int retryCount = getRetryCount();
		T result = null;
		for (int i = 1; i <= retryCount; i++) {
			try {
				result = handler.handle();
				break;
			} catch (Exception e) {
				if (retryCount == i) {
					throw e;
				} else {
					String causePoint = null;
					for (StackTraceElement ste : e.getStackTrace()) {
						if (ste.getClassName().startsWith("cs.")) {
							causePoint = ste.toString();
							break;
						}
					}
					System.err.printf("Worry & Retry!(%d of %d) [%s]%s Message: %s\n", i,
							retryCount,
							causePoint,
							e.getClass().getName(),
							e.getLocalizedMessage());
				}
			}
		}
		return result;
	}
}
